with
  text_io,
  ada.integer_text_io,
  ada.long_float_text_io,
  terms,
  nlp;

use
  text_io,
  ada.integer_text_io,
  ada.long_float_text_io;

procedure rosenbrock is

  optimal                        : boolean;
  n,maxls,minor,maxminor,maxlist : integer;
  lstol,maxnorm,f                : long_float;
  x,g                            : terms.fmat1_ptr;

  procedure objective (n : in integer;
                       f : out long_float;
                       x : in terms.fmat1_ptr;
                       g : in terms.fmat1_ptr) is

  begin
    -- objective
    f := 100.0 * (x(1) ** 2 - x(2)) ** 2 + (1.0 - x(1)) ** 2;

    -- gradient
    g(1) := 400.0 * (x(1) ** 2 - x(2)) * x(1) - 2.0 * (1.0 - x(1));
    g(2) := -200.0 * (x(1) ** 2 - x(2));
  end objective;

  package qn is new nlp.qnewton(objective);

begin
  -- F : The smooth nonlinear real-valued function F(X).

  -- G : The gradient vector of F(X).

  -- OBJECTIVE : A procedure that evaluates function value and
  -- gradient vector.

  -- MINOR : The number of iterations required.

  -- OPTIMAL : The state at finishing. It assumes true if
  -- success, false otherwise.

  -- MAXMINOR : The maximum number of iterations allowed. It
  -- must be greater that zero. It is usually 10 * N.
  maxminor := 100;

  -- MAXNORM : The maximum value allowed for the infinity norm of
  -- the gradient. The iterations terminate if the infinity norm
  -- of the gradient becomes less than or equal to MAXNORM.
  maxnorm := 1.0E-06;

  -- MAXLS : The maximum number of cubic interpolations allowed.
  -- It must be greater that one. It is usually 20.
  maxls := 20;

  -- LSTOL : Controls the accuracy of the search. It must lie in
  -- the range 0.0 <= LSTOL < 1.0. Decreasing this value tends
  -- to increase the accuracy of the search. It is usually 0.1.
  lstol := 0.1;

  -- MAXLIST : The number of BFGS corrections kept. These
  -- corrections are stored separately on a doubly linked
  -- circular list. Used only with the limited memory BFGS
  -- method. It is usually between 3 and 20.
  maxlist := 3;

  -- N : The number of variables.
  n := 2;

  -- initiate x
  x := new terms.fmat1 (1 .. n);
  for k in 1 .. n loop
    x(k) := 0.0;
  end loop;

  -- initiate g
  g := new terms.fmat1 (1 .. n);
  for k in 1 .. n loop
    g(k) := 0.0;
  end loop;

  -- X : The vector of variables (starting point).
  x(1) := -1.0;
  x(2) := -1.0;

  qn.lbfgs(maxlist,n,maxminor,maxls,
           minor,
           maxnorm,lstol,
           f,
           x,g,
           optimal);

  if optimal then
    put("x(1) = ");
    put(x(1),2,8,3);
    new_line;

    put("x(2) = ");
    put(x(2),2,8,3);
    new_line;
  else
    put("Not optimal");
    new_line;
  end if;

  put("Iterations = ");
  put(minor,5);
  new_line;
end rosenbrock;
